'''
Module provider for EUserv

Author: Matthias Schoettle (@mschoettle), 2019

EUserv API Docs: https://support.euserv.com/api-doc/
'''
from __future__ import absolute_import

import json
import logging
import requests

from lexicon.providers.base import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ['euserv.com']

# Success response code
RC_SUCCESS = 100

API_ENDPOINT = 'https://support.euserv.com/'

# Product group ID for domains
PRODUCT_ID_DOMAIN = 1


def provider_parser(subparser):
    '''Configure provider parser for Euserv'''
    subparser.add_argument(
        '--auth-username', help='specify email address for authentication')
    subparser.add_argument(
        '--auth-password', help='specify password for authentication')


class Provider(BaseProvider):
    '''Provider class for Euserv'''
    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.api_endpoint = API_ENDPOINT
        # Order ID for the domain
        self.order_id = None
        self.domain_id = None
        self.session_id = None

    def _authenticate(self):
        if not (self._get_provider_option('auth_username')
                and self._get_provider_option('auth_password')):
            raise Exception('username and password must be specified, add --help for details')

        # Get a session ID first.
        LOGGER.info('Getting Session ID...')
        response = self._get()
        self.session_id = response['result']['sess_id']['value']

        LOGGER.info('Logging in...')
        auth_response = self._get(query_params={
            'subaction': 'login',
            'email': self._get_provider_option('auth_username'),
            'password': self._get_provider_option('auth_password')
            })

        # Find the contract number of the given domain
        orders = auth_response['result']['orders']

        for order in orders:
            if int(order['pg_id']['value']) == PRODUCT_ID_DOMAIN:
                # The description contains the description of the product itself
                # and in a second line the domain name
                order_description = order['ord_description']['value'].split('\n')

                if order_description[1] == self.domain:
                    self.order_id = order['ord_no']['value']
                    break

        if self.order_id is None:
            raise Exception('Order for domain not found')

        # Select the order for the given domain so we can use the DNS actions
        LOGGER.info('Choosing order %s', self.order_id)
        self._get(query_params={
            'subaction': 'choose_order',
            'ord_no': self.order_id
            })

        # Retrieve domain ID
        LOGGER.info('Retrieving DNS records to find domain id for %s...', self.domain)
        domains = self._get(query_params={'subaction': 'kc2_domain_dns_get_records'})

        for domain in domains['result']['domains']:
            if domain['dom_domain']['value'] == self.domain:
                self.domain_id = domain['dom_id']['value']
                break

        if self.domain_id is None:
            raise Exception('Domain not found in DNS records')

    def _create_record(self, rtype, name, content):
        records = self._list_records(rtype, name, content)

        if records:
            LOGGER.debug('record already exists: %s', records[0])
            return True

        query_params = {
            'subaction': 'kc2_domain_dns_set',
            'dom_id': self.domain_id,
            'type': rtype,
            'content': content
        }

        if name:
            query_params['subdomain'] = self._subdomain_name(name)

        self._add_ttl(query_params)
        self._add_priority(query_params)

        response = self._get(query_params=query_params)
        LOGGER.debug('create_record response: %s', response)

        return True

    def _list_records(self, rtype=None, name=None, content=None):
        LOGGER.info('Listing records for type=%s, name=%s, content=%s', rtype, name, content)

        query_params = {
            'subaction': 'kc2_domain_dns_get_records',
            'dns_records_load_only_for_dom_id': self.domain_id
        }

        if rtype:
            query_params['dns_records_load_type'] = rtype

        if name:
            query_params['dns_records_load_subdomain'] = self._subdomain_name(name)

        if content:
            query_params['dns_records_load_content'] = content

        payload = self._get(query_params=query_params)

        response = payload['result']['domains'][0]

        LOGGER.debug('list_records raw response: %s', response)

        records = []

        if 'dns_records' in response:
            for record in response['dns_records']:
                processed_record = {
                    'type': record['type']['value'],
                    'name': record['name']['value'],
                    'ttl': record['ttl']['value'],
                    'content': record['content']['value'],
                    'id': record['id']['value'],
                    'priority': record['prio']['value']
                }

                records.append(processed_record)

        return records

    def _update_record(self, identifier, rtype=None, name=None, content=None):
        if identifier is None:
            LOGGER.info('No identifier for record provided, trying to find it...')
            identifier = self._find_record_identifier(rtype, name, None)

        query_params = {
            'subaction': 'kc2_domain_dns_set',
            'dns_record_id': identifier
        }

        if rtype:
            query_params['type'] = rtype

        if name:
            query_params['subdomain'] = self._subdomain_name(name)

        if content:
            query_params['content'] = content

        self._add_ttl(query_params)
        self._add_priority(query_params)

        response = self._get(query_params=query_params)
        LOGGER.debug('update_record: %s', response)

        return True

    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        record_ids = []

        if identifier is None:
            LOGGER.info('No identifier for record provided, trying to find it...')
            records = self._list_records(rtype, name, content)
            record_ids = [record['id'] for record in records]
        else:
            record_ids.append(identifier)

        LOGGER.debug('record IDs to be deleted: %s', record_ids)

        for record_id in record_ids:
            query_params = {
                'subaction': 'kc2_domain_dns_remove',
                'dns_record_id': record_id
            }

            response = response = self._get(query_params=query_params)
            LOGGER.debug('delete_record: %s', response)

        return True

    # Helpers
    def _request(self, action='GET', url='/', data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}

        query_params['method'] = 'json'

        if self.session_id:
            query_params['sess_id'] = self.session_id

        response = requests.request(action, self.api_endpoint, params=query_params,
                                    data=json.dumps(data),
                                    headers={'Content-Type': 'application/json; charset=utf-8'}
                                    )
        # if the request fails for any reason, throw an error.
        response.raise_for_status()

        response_json = response.json()

        if int(response_json['code']) != RC_SUCCESS:
            raise Exception('Error {0} in request: {1}'.format(response_json['code'],
                                                               response_json['message']))

        return response_json

    # Adds TTL parameter if passed as argument to lexicon.
    def _add_ttl(self, data):
        if self._get_lexicon_option('ttl'):
            data['ttl'] = int(self._get_lexicon_option('ttl'))

    # Adds priority parameter if passed as argument to lexicon.
    def _add_priority(self, data):
        if self._get_lexicon_option('priority'):
            data['prio'] = int(self._get_lexicon_option('priority'))

    # Find identifier of a record with the given properties.
    def _find_record_identifier(self, rtype, name, content):
        records = self._list_records(rtype, name, content)

        LOGGER.debug('records: %s', records)

        if len(records) == 1:
            return records[0]['id']

        raise Exception('No record identifier found')

    # Get the subdomain name only for the given name.
    # This provider automatically suffixes the name with the domain name.
    def _subdomain_name(self, name):
        subdomain = self._full_name(name)
        domain_suffix = '.' + self.domain

        # Remove domain name since it will be automatically added by the provider
        if subdomain.endswith(domain_suffix):
            subdomain = subdomain[:len(subdomain)-len(domain_suffix)]

        return subdomain



"""Integration tests for INWX"""
from unittest import TestCase

from lexicon.tests.providers.integration_tests import IntegrationTestsV2


class InwxProviderTests(TestCase, IntegrationTestsV2):
    """TestCase for INWX"""
    provider_name = 'inwx'
    domain = 'lexicon-testrunner.com'

    def _test_parameters_overrides(self):
        return {
            'endpoint': 'https://api.ote.domrobot.com/xmlrpc/'
        }



"""Integration tests for DigitalOcean"""
from unittest import TestCase

import pytest
from lexicon.tests.providers.integration_tests import IntegrationTestsV2


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from define_tests.TheTests
class DinahostingProviderTests(TestCase, IntegrationTestsV2):
    """Integration tests for Dinahosting provider"""
    provider_name = 'dinahosting'
    domain = 'itslikethesesweet.info'

    def _filter_headers(self):
        return ['Authorization', 'Set-Cookie']

    @pytest.mark.skip(reason="API does not expose record ttl")
    def test_provider_when_calling_list_records_after_setting_ttl(self):
        return



"""Integration tests for Conoha"""
from unittest import TestCase

from lexicon.tests.providers.integration_tests import IntegrationTestsV2


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from define_tests.TheTests
class ConohaProviderTests(TestCase, IntegrationTestsV2):
    """TestCase for Conoha"""
    provider_name = 'conoha'
    domain = 'narusejun.com'

    def _test_parameters_overrides(self):
        return {'region': 'tyo1'}

    def _test_fallback_fn(self):
        return lambda x: None if x in ('priority') else 'placeholder_' + x

    def _filter_post_data_parameters(self):
        return ['auth']

    def _filter_headers(self):
        return ['X-Auth-Token']



"""Unit tests for the Lexicon config mechanism"""
# pylint: disable=redefined-outer-name,missing-docstring
from lexicon.config import ConfigResolver, ConfigSource
from lexicon.parser import generate_cli_main_parser


def test_environment_resolution(monkeypatch):
    monkeypatch.setenv('LEXICON_DELEGATED', 'TEST1')
    monkeypatch.setenv('LEXICON_CLOUDFLARE_TOKEN', 'TEST2')
    monkeypatch.setenv('LEXICON_CLOUDFLARE_AUTH_USERNAME', 'TEST3')

    config = ConfigResolver().with_env()

    assert config.resolve('lexicon:delegated') == 'TEST1'
    assert config.resolve('lexicon:cloudflare:auth_token') == 'TEST2'
    assert config.resolve('lexicon:cloudflare:auth_username') == 'TEST3'
    assert config.resolve('lexicon:nonexistent') is None


def test_argparse_resolution():
    parser = generate_cli_main_parser()
    data = parser.parse_args(['--delegated', 'TEST1', 'cloudflare',
                              'create', 'example.com', 'TXT', '--auth-token', 'TEST2'])

    config = ConfigResolver().with_args(data)

    assert config.resolve('lexicon:delegated') == 'TEST1'
    assert config.resolve('lexicon:cloudflare:auth_token') == 'TEST2'
    assert config.resolve('lexicon:nonexistent') is None


def test_dict_resolution():
    dict_object = {
        'delegated': 'TEST1',
        'cloudflare': {
            'auth_token': 'TEST2'
        }
    }

    config = ConfigResolver().with_dict(dict_object)

    assert config.resolve('lexicon:delegated') == 'TEST1'
    assert config.resolve('lexicon:cloudflare:auth_token') == 'TEST2'
    assert config.resolve('lexicon:nonexistent') is None


def test_config_lexicon_file_resolution(tmpdir):
    lexicon_file = tmpdir.join('lexicon.yml')
    lexicon_file.write('delegated: TEST1\ncloudflare:\n  auth_token: TEST2')

    config = ConfigResolver().with_config_file(str(lexicon_file))

    assert config.resolve('lexicon:delegated') == 'TEST1'
    assert config.resolve('lexicon:cloudflare:auth_token') == 'TEST2'
    assert config.resolve('lexicon:nonexistent') is None


def test_provider_config_lexicon_file_resolution(tmpdir):
    provider_file = tmpdir.join('lexicon_cloudflare.yml')
    provider_file.write('auth_token: TEST2')

    config = ConfigResolver().with_provider_config_file(
        'cloudflare', str(provider_file))

    assert config.resolve('lexicon:cloudflare:auth_token') == 'TEST2'
    assert config.resolve('lexicon:nonexistent') is None


def test_provider_config_dir_resolution(tmpdir):
    lexicon_file = tmpdir.join('lexicon.yml')
    provider_file = tmpdir.join('lexicon_cloudflare.yml')
    lexicon_file.write('delegated: TEST1\ncloudflare:\n  auth_token: TEST2')
    provider_file.write('auth_username: TEST3')

    config = ConfigResolver().with_config_dir(str(tmpdir))

    assert config.resolve('lexicon:delegated') == 'TEST1'
    assert config.resolve('lexicon:cloudflare:auth_token') == 'TEST2'
    assert config.resolve('lexicon:cloudflare:auth_username') == 'TEST3'
    assert config.resolve('lexicon:nonexistent') is None


def test_generic_config_feeder_resolution():
    class GenericConfigSource(ConfigSource):  # pylint: disable=too-few-public-methods
        def resolve(self, config_key):
            return 'TEST1'

    config = ConfigResolver().with_config_source(GenericConfigSource())

    assert config.resolve('lexicon:cloudflare:auth_username') == 'TEST1'
    assert config.resolve('lexicon:nonexistent') == 'TEST1'


def test_legacy_dict_config_resolution():
    legacy_config = {
        'delegated': 'TEST1',
        'auth_token': 'TEST2',
        'provider_name': 'cloudflare'
    }

    config = ConfigResolver().with_legacy_dict(legacy_config)

    assert config.resolve('lexicon:delegated') == 'TEST1'
    assert config.resolve('lexicon:cloudflare:auth_token') == 'TEST2'
    assert config.resolve('lexicon:auth_token') is None
    assert config.resolve('lexicon:nonexistent') is None


def test_prioritized_resolution(tmpdir, monkeypatch):
    lexicon_file = tmpdir.join('lexicon.yml')
    lexicon_file.write('cloudflare:\n  auth_token: TEST1')

    monkeypatch.setenv('LEXICON_CLOUDFLARE_AUTH_TOKEN', 'TEST2')

    assert ConfigResolver().with_config_file(str(lexicon_file)).with_env().resolve(
        'lexicon:cloudflare:auth_token') == 'TEST1'
    assert ConfigResolver().with_env().with_config_file(
        str(lexicon_file)).resolve('lexicon:cloudflare:auth_token') == 'TEST2'



"""Module provider for UKFast's SafeDNS"""
from __future__ import absolute_import
import json
import logging

import requests
from lexicon.providers.base import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ['ukfast.net']

def provider_parser(subparser):
    """Configure provider parser for SafeDNS"""
    subparser.description = '''
        SafeDNS provider requires an API key in all interactions.
        You can generate one for your account on the following URL:
        https://my.ukfast.co.uk/applications/index.php'''
    subparser.add_argument('--auth-token', help='specify the API key to authenticate with')

class Provider(BaseProvider):
    """Provider SafeDNS implementation of Lexicon Provider interface."""
    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = 'https://api.ukfast.io/safedns/v1'

    def _authenticate(self):
        try:
            self._get('/zones/{0}'.format(self.domain))
            self.domain_id = self.domain
        except requests.exceptions.HTTPError as err:
            if err.response.status_code == 404:
                raise Exception('No domain found')
            raise err

    # List all records. Return an empty list if no records found.
    # type, name and content are used to filter records.
    def _list_records(self, rtype=None, name=None, content=None):
        url = '/zones/{0}/records'.format(self.domain_id)
        records = []
        payload = {}

        # Handle pagination and iteratively collect all records
        next_url = url
        while next_url is not None:
            payload = self._get(next_url)
            if 'meta' in payload \
                    and 'pagination' in payload['meta'] \
                    and 'links' in payload['meta']['pagination'] \
                    and 'next' in payload['meta']['pagination']['links']:
                next_url = payload['meta']['pagination']['links']['next']
            else:
                next_url = None

            # Assign the returned attributes to the keys that lexicon expects
            for record in payload['data']:
                record = self._clean_TXT_record(record)
                processed_record = {
                    'id': record['id'],
                    'name': record['name'],
                    'type': record['type'],
                    'content': record['content'],
                    'updated_at': record['updated_at'],
                    'ttl': record['ttl'],
                    'priority': record['priority']
                }
                records.append(processed_record)

        # This is filtering logic to return only the record which matches what has
        # been passed in to the method
        if rtype:
            records = [
                record for record in records
                if record['type'] == rtype]
        if name:
            records = [
                record for record in records
                # DNS should not be case-sensitive, so perform lower-case comparison
                if record['name'].lower() == self._full_name(name).lower()]
        if content:
            records = [
                record for record in records
                if record['content'].lower() == content.lower()]

        LOGGER.debug('list_records: %s', records)
        return records

    def _create_record(self, rtype, name, content):

        # Check whether the record already exists with the same  rtype, name & content.
        # If so, claim to have added the record, but dont't do anything.
        records = self._list_records(rtype, name, content)
        if records:
            LOGGER.debug('create_record: (ignored, duplicate record): %s', records[0]['id'])
            return True

        # Make sure TXT records are wrapped in quotes
        if content:
            content = self._add_quotes(rtype, content)

        data = {
            'name': self._full_name(name),
            'type': rtype,
            'content': content
        }

        self._post('/zones/{0}/records'.format(self.domain), data)
        LOGGER.debug('create_record: %s', True)
        return True

    def _update_record(self, identifier, rtype=None, name=None, content=None):

        # Make sure the update won't cause a duplicate entry. If it will, fail silently
        records = self._list_records(rtype, name, content)

        if not identifier:
            records = self._list_records(rtype, name)
            if len(records) == 1:
                identifier = records[0]['id']
            elif len(records) > 1:
                identifier = records[0]['id']
                LOGGER.warning(
                    'Warning, multiple records found for given parameters, '
                    'only first one will be updated: %s', records)
            else:
                raise Exception('Record identifier could not be found')

        # Make sure TXT records are wrapped in quotes
        if content:
            content = self._add_quotes(rtype, content)

        data = {}
        if name:
            data['name'] = self._full_name(name)
        if rtype:
            data['type'] = rtype
        if content:
            data['content'] = content

        self._patch(
            '/zones/{0}/records/{1}'.format(self.domain, identifier), data)

        LOGGER.debug('update_record: %s', True)
        return True

    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):

        delete_record_ids = []

        # If we've not been given an identifier, search for matching records.
        # NOTE, this could cause multiple records to be removed.
        if not identifier:
            records = self._list_records(rtype, name, content)
            delete_record_ids = [record['id'] for record in records]
        else:
            delete_record_ids.append(identifier)

        LOGGER.debug('delete_records: %s', delete_record_ids)

        for delete_record_id in delete_record_ids:
            self._delete(
                '/zones/{0}/records/{1}'.format(self.domain, delete_record_id))

        LOGGER.debug('delete_record: %s', True)
        return True

    def _request(self, action='GET', url='/', data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        default_headers = {
            'Accept': 'application/json',
            'Content-Type': 'application/json',
            'Authorization': '{0}'.format(self._get_provider_option('auth_token'))
        }
        if not url.startswith(self.api_endpoint):
            url = self.api_endpoint + url

        response = requests.request(action, url, params=query_params,
                                    data=json.dumps(data),
                                    headers=default_headers)

        # Sort this out to work properly, writing errors to the correct place
        if 'content' in response:
            resp = json.loads(response.content)
            if 'errors' in resp:
                for error in resp['errors']:
                    print("ERROR: " + error['detail'])

        # If the request fails for any reason, throw an error.
        response.raise_for_status()

        # There is no JSON returned when calling with DELETE
        if action != 'DELETE':
            return response.json()

        return True

    # The content of TXT entries must be quoted. This static method ensures that.
    @staticmethod
    def _add_quotes(rtype, content):
        if rtype == 'TXT':
            if not content.startswith('"') and not content.endswith('"'):
                return '"{0}"'.format(content)
        return content



"""Test for one implementation of the interface"""
from unittest import TestCase
import pytest
from lexicon.tests.providers.integration_tests import IntegrationTestsV2

# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from integration_tests.IntegrationTests
class SafednsProviderTests(TestCase, IntegrationTestsV2):
    """Integration tests for SafeDNS provider"""
    provider_name = 'safedns'
    domain = 'lexicon.tests'

    def _filter_headers(self):
        return ['Authorization']

    @pytest.mark.skip(reason="Record-level TTLs are not supported by this provider")
    def test_provider_when_calling_list_records_after_setting_ttl(self):
        return

    @pytest.mark.skip(reason="CNAME requires FQDN for this provider")
    def test_provider_when_calling_create_record_for_CNAME_with_valid_name_and_content(self):
        return



"""Module provider for Softlayer"""
from __future__ import absolute_import
import logging

from lexicon.providers.base import Provider as BaseProvider

try:
    import SoftLayer
except ImportError:
    pass

LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ['softlayer.com']


def provider_parser(subparser):
    """Generate a provider parser for Softlayer"""
    subparser.add_argument(
        "--auth-username", help="specify username for authentication")
    subparser.add_argument(
        "--auth-api-key", help="specify API private key for authentication")


class Provider(BaseProvider):
    """Provider class for Softlayer"""
    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None

        username = self._get_provider_option('auth_username')
        api_key = self._get_provider_option('auth_api_key')

        if not username or not api_key:
            raise Exception("No username and/or api key was specified")

        sl_client = SoftLayer.create_client_from_env(
            username=username, api_key=api_key)
        self.sl_dns = SoftLayer.managers.dns.DNSManager(sl_client)

    # Authenticate against provider,
    # Make any requests required to get the domain's id for this provider,
    # so it can be used in subsequent calls.
    # Should throw an error if authentication fails for any reason,
    # of if the domain does not exist.
    def _authenticate(self):
        domain = self.domain

        payload = self.sl_dns.resolve_ids(domain)

        if not payload:
            raise Exception('No domain found')
        if len(payload) > 1:
            raise Exception('Too many domains found. This should not happen')

        LOGGER.debug('domain id: %s', payload[0])
        self.domain_id = payload[0]

    # Create record. If record already exists with the same content, do nothing

    def _create_record(self, rtype, name, content):
        records = self._list_records(rtype, name, content)
        if records:
            # Nothing to do, record already exists
            LOGGER.debug('create_record: already exists')
            return True

        name = self._relative_name(name)
        ttl = self._get_lexicon_option('ttl')
        payload = self.sl_dns.create_record(
            self.domain_id, name, rtype, content, ttl)

        LOGGER.debug('create_record: %s', payload)
        return True

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.

    def _list_records(self, rtype=None, name=None, content=None):
        ttl = None
        if name:
            name = self._relative_name(name)

        payload = self.sl_dns.get_records(
            self.domain_id, ttl, content, name, rtype)

        records = []
        for record in payload:
            processed_record = {
                'type': record['type'].upper(),
                'name': self._full_name(record['host']),
                'ttl': record['ttl'],
                'content': record['data'],
                'id': record['id']
            }
            records.append(processed_record)

        LOGGER.debug('list_records: %s', records)
        return records

    # Update a record.
    # If an identifier is specified, use it, otherwise do a lookup using type and name.

    def _update_record(self, identifier=None, rtype=None, name=None, content=None):
        if not identifier:
            records = self._list_records(rtype, name)
            if len(records) == 1:
                identifier = records[0]['id']
            else:
                raise Exception('Record identifier could not be found.')

        record = {'id': identifier}
        if rtype:
            record['type'] = rtype
        if name:
            record['host'] = self._relative_name(name)
        if content:
            record['data'] = content
        if self._get_lexicon_option('ttl'):
            record['ttl'] = self._get_lexicon_option('ttl')

        self.sl_dns.edit_record(record)

        LOGGER.debug('update_record: %s', record)
        return True

    # Delete an existing record.
    # If record does not exist, do nothing.
    # If an identifier is specified, use it, otherwise do a lookup using type, name and content.

    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        delete_record_id = []
        if not identifier:
            records = self._list_records(rtype, name, content)
            delete_record_id = [record['id'] for record in records]
        else:
            delete_record_id.append(identifier)

        LOGGER.debug('delete_records: %s', delete_record_id)

        for record_id in delete_record_id:
            self.sl_dns.delete_record(record_id)

        LOGGER.debug('delete_record: %s', True)
        return True

    def _request(self, action='GET', url='/', data=None, query_params=None):
        # Helper _request is not used in Softlayer
        pass



"""Module provider for Hetzner"""
from __future__ import absolute_import
from __future__ import unicode_literals

import json
import logging
import requests

from lexicon.providers.base import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)
NAMESERVER_DOMAINS = ['ns.hetzner.com']


def provider_parser(subparser):
    """Configure a provider parser for Hetzner"""
    subparser.add_argument('--auth-token', help='Specify Hetzner DNS API token')


class Provider(BaseProvider):
    """
    Implements the Hetzner DNS Provider using the service https://dns.hetzner.com.
    It neither works for their web hosting offer "konsoleH" nor for the discontinued
    "Domain Robot" which are supported by the ``lexicon.providers.hetzner_legacy`` module.
    """

    API_VERSION = '1.0'

    """Provider class for Cloudflare"""

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = 'https://dns.hetzner.com/api/v1'

    def _authenticate(self):
        provider = self._get_zone_by_domain(self.domain)
        self.domain_id = provider['id']

    def _create_record(self, rtype, name, content):
        """
        Creates a DNS record, if a record with type, name and content exits, do nothing
        :returns: A json string containing the resulting record
        :rtype: str
        """

        data = {
            # hetzner needs the FQDN name if it does not belong to the managed domain itself
            'name': self._get_record_name(self.domain, name),
            'type': rtype,
            'value': content,
            'zone_id': self.domain_id
        }
        if self._get_lexicon_option('ttl'):
            data['ttl'] = self._get_lexicon_option('ttl')

        records = self._list_records(rtype=rtype, name=name, content=content)
        if len(records) >= 1:
            for record in records:
                LOGGER.warning(
                    "Duplicate record %s %s %s with id %s",
                    rtype, name, content, record['id']
                )
            return True
        self._post('/records', data)
        return True

    def _list_records(self, rtype=None, name=None, content=None):
        """
        List all records, filterable by type, name and content
        :rtype: list
        :returns: list of records, might be empty
        """
        filter_obj = {
            'per_page': 100,
            'zone_id': self.domain_id
        }
        payload = self._get('/records', filter_obj)
        records = map(self._hetzner_record_to_lexicon_record, payload['records'])
        filtered_records = self._filter_records(
            records, rtype, name if name is not None else None, content
        )

        return filtered_records

    def _filter_records(self, records, rtype=None, name=None, content=None):
        return [
            record for record in records
            if (rtype is None or record['type'] == rtype) and
            (name is None or record['name'] == self._full_name(name)) and
            (content is None or record['content'] == content)
        ]

    def _update_record(self, identifier, rtype=None, name=None, content=None):
        """
        Create or update a record.
        :rtype: bool
        :returns: Boolean indicating success
        """
        data = {
            'type': rtype,
            'name': self._get_record_name(self.domain, name),
            'value': content,
            'zone_id': self.domain_id
        }
        if self._get_lexicon_option('ttl'):
            data['ttl'] = self._get_lexicon_option('ttl')
        update_identifier = identifier
        if update_identifier is None:
            records = self._list_records(rtype, name)
            if len(records) == 1:
                update_identifier = records[0]['id']
            elif len(records) < 1:
                raise Exception('No records found matching type, name and content - won\'t update')
            else:
                raise Exception(
                    'Multiple records found matching type, name and content - won\'t update'
                )
        self._put('/records/{0}'.format(update_identifier), data)
        return True

    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        """
        Delete an existing record. If record does not exist, do nothing.
        :rtype: bool
        :returns: Boolean indicating success
        """
        delete_record_ids = []
        if identifier is None:
            records = self._list_records(rtype, name, content)
            delete_record_ids = [record['id'] for record in records]
        else:
            delete_record_ids.append(identifier)

        for record_id in delete_record_ids:
            self._delete('/records/{0}'.format(record_id))
        return True

    # Helpers
    def _request(self, action='GET', url='/', data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        response = requests.request(action, self.api_endpoint + url, params=query_params,
                                    data=json.dumps(data),
                                    headers={
                                        'Auth-API-Token': self._get_provider_option('auth_token'),
                                        'Content-Type': 'application/json'
                                    })
        # if the request fails for any reason, throw an error.
        response.raise_for_status()
        return response.json()

    def _get_zone_by_domain(self, domain):
        """
        Requests all dns zones from your Hetzner account and searches for a specific
        one to determine the ID of it
        :param domain: Name of domain for which dns zone should be searched
        :rtype: dict
        :return: The dictionary of the zone with ``domain`` in the 'name' key
        :raises Exception: If no zone was found
        :raises KeyError, ValueError: If the response is malformed
        :raises urllib.error.HttpError: If request to /zones did not return 200
        """
        payload = self._get('/zones')
        zones = payload['zones']
        for zone in zones:
            if zone['name'] == domain:
                return zone
        raise Exception("No zone was found in account matching {0}".format(domain))

    def _get_record_name(self, domain, record_name):
        """
        Get the name attribute appropriate for hetzner api. This means it's the name
        without domain name if record name ends with managed domain name else a fqdn
        :param domain: Name of domain for which dns zone should be searched
        :param record_name: The record name to convert
        :rtype: str
        :return: The record name in an appropriate format for hetzner api
        """
        if record_name.rstrip('.').endswith(domain):
            record_name = self._relative_name(record_name)
        return record_name

    @staticmethod
    def _pretty_json(data):
        return json.dumps(data, sort_keys=True, indent=4, separators=(',', ': '))

    def _hetzner_record_to_lexicon_record(self, hetzner_record):
        lexicon_record = {
            'id': hetzner_record['id'],
            'name': self._full_name(hetzner_record['name']),
            'content': hetzner_record['value'],
            'type': hetzner_record['type']
        }
        if 'ttl' in hetzner_record:
            lexicon_record['ttl'] = hetzner_record['ttl']
        return lexicon_record



"""Parsers definition for the Lexicon command-line interface"""
import argparse
import importlib
import os

from lexicon import discovery


def generate_base_provider_parser():
    """Function that generates the base provider to be used by all dns providers."""
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument('action', help='specify the action to take', default='list',
                        choices=['create', 'list', 'update', 'delete'])
    parser.add_argument(
        'domain', help='specify the domain, supports subdomains as well')
    parser.add_argument('type', help='specify the entry type', default='TXT',
                        choices=['A', 'AAAA', 'CNAME', 'MX', 'NS', 'SOA', 'TXT', 'SRV', 'LOC'])

    parser.add_argument('--name', help='specify the record name')
    parser.add_argument('--content', help='specify the record content')
    parser.add_argument('--ttl', type=int,
                        help='specify the record time-to-live')
    parser.add_argument('--priority', help='specify the record priority')
    parser.add_argument(
        '--identifier', help='specify the record for update or delete actions')
    parser.add_argument('--log_level', help='specify the log level', default='ERROR',
                        choices=['CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG', 'NOTSET'])
    parser.add_argument('--output',
                        help=('specify the type of output: by default a formatted table (TABLE), '
                              'a formatted table without header (TABLE-NO-HEADER), '
                              'a JSON string (JSON) or no output (QUIET)'),
                        default='TABLE', choices=['TABLE', 'TABLE-NO-HEADER', 'JSON', 'QUIET'])
    return parser


def generate_cli_main_parser():
    """Using all providers available, generate a parser that will be used by Lexicon CLI"""
    parser = argparse.ArgumentParser(
        description='Create, Update, Delete, List DNS entries')

    parser.add_argument('--version', help='show the current version of lexicon',
                        action='version', version='%(prog)s {0}'
                        .format(discovery.lexicon_version()))
    parser.add_argument('--delegated', help='specify the delegated domain')
    parser.add_argument('--config-dir', default=os.getcwd(),
                        help='specify the directory where to search lexicon.yml and '
                             'lexicon_[provider].yml configuration files '
                             '(default: current directory).')
    subparsers = parser.add_subparsers(
        dest='provider_name', help='specify the DNS provider to use')
    subparsers.required = True

    for provider, available in discovery.find_providers().items():
        provider_module = importlib.import_module(
            'lexicon.providers.' + provider)
        provider_parser = getattr(provider_module, 'provider_parser')

        subparser = subparsers.add_parser(provider, help='{0} provider'.format(provider),
                                          parents=[generate_base_provider_parser()])
        provider_parser(subparser)

        if not available:
            subparser.epilog = ('WARNING: some required dependencies for this provider are not '
                                'installed. Please install lexicon[{0}] first before using it.'
                                .format(provider))

    return parser



"""Integration tests for LuaDNS"""
from unittest import TestCase

import pytest
from lexicon.tests.providers.integration_tests import IntegrationTestsV2


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from define_tests.TheTests
class LuaDNSProviderTests(TestCase, IntegrationTestsV2):
    """TestCase for LuaDNS"""
    provider_name = 'luadns'
    domain = 'capsulecd.com'

    def _filter_headers(self):
        return ['Authorization']

    @pytest.mark.skip(reason="CNAME requires FQDN for this provider")
    def test_provider_when_calling_create_record_for_CNAME_with_valid_name_and_content(self):
        return

    # TODO: enable the skipped tests
    @pytest.mark.skip(reason="new test, missing recording")
    def test_provider_when_calling_update_record_should_modify_record_name_specified(self):
        return



"""Integration tests for CloudNS"""
from unittest import TestCase

from lexicon.tests.providers.integration_tests import IntegrationTestsV2


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from define_tests.TheTests
class CloudnsProviderTests(TestCase, IntegrationTestsV2):
    """TestCase for CloudNS"""
    provider_name = 'cloudns'
    domain = 'api-example.com'

    def _filter_query_parameters(self):
        return ['auth-id', 'sub-auth-id', 'sub-auth-user', 'auth-password']

    def _filter_post_data_parameters(self):
        return ['auth-id', 'sub-auth-id', 'sub-auth-user', 'auth-password']



"""Integration tests for DirectAdmin"""
from unittest import TestCase
from lexicon.tests.providers.integration_tests import IntegrationTestsV2


# Hook into testing framework by inheriting unittest.TestCase and reuse the
# tests which *each and every* implementation of the interface must pass, by
# inheritance from integration_tests.IntegrationTests
class DirectAdminProviderTests(TestCase, IntegrationTestsV2):
    """Integration tests for DirectAdmin provider"""
    provider_name = 'directadmin'
    domain = 'example.com'
    endpoint = 'http://localhost'

    def _filter_post_data_parameters(self):
        return []

    def _filter_headers(self):
        return ['Authorization']

    def _filter_query_parameters(self):
        return []

    def _filter_response(self, response):
        """See `IntegrationTests._filter_response` for more information on how
           to filter the provider response."""
        return response

    def _test_parameters_overrides(self):
        return {'endpoint': self.endpoint}



"""Module provider for Linode V4"""
from __future__ import absolute_import
import json
import logging

import requests
from lexicon.providers.base import Provider as BaseProvider


LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ['linode.com']


def provider_parser(subparser):
    """Configure provider parser for Linode V4"""
    subparser.add_argument(
        "--auth-token", help="specify api key for authentication")


class Provider(BaseProvider):
    """Provider class for Linode V4"""
    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = 'https://api.linode.com/v4/'

    def _authenticate(self):
        self.domain_id = None
        payload = self._get('domains', query_params={
            'filter': {
                'domain': self.domain.lower()
            }
        })
        if payload['data']:
            self.domain_id = payload['data'][0]['id']
        if self.domain_id is None:
            raise Exception('Domain not found')

    def _create_record(self, rtype, name, content):
        if not self._list_records(rtype, name, content):
            if name:
                name = self._relative_name(name).lower()

            self._post('domains/{0}/records'.format(self.domain_id), data={
                'name': name,
                'type': rtype,
                'target': content,
                'ttl_sec': 0
            })

        return True

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def _list_records(self, rtype=None, name=None, content=None):
        resources_url = "domains/{0}/records".format(self.domain_id)

        if name:
            name = self._relative_name(name).lower()

        processed_records = []

        payload = self._get(resources_url)
        for page in range(1, payload['pages'] + 1, 1):
            if page > 1:
                payload = self._get(resources_url, query_params={
                    'page': page
                })

            resource_list = payload['data']
            if rtype:
                resource_list = [
                    resource for resource in resource_list if resource['type'] == rtype]
            if name:
                resource_list = [resource for resource in resource_list if self._relative_name(
                    resource['name']).lower() == name]
            if content:
                resource_list = [
                    resource for resource in resource_list if resource['target'] == content]

            for resource in resource_list:
                processed_records.append({
                    'id': resource['id'],
                    'type': resource['type'],
                    'name': self._full_name(resource['name']).lower(),
                    'ttl': resource['ttl_sec'],
                    'content': resource['target']
                })

        LOGGER.debug('list_records: %s', processed_records)
        return processed_records

    # Create or update a record.
    def _update_record(self, identifier, rtype=None, name=None, content=None):
        if not identifier:
            resources = self._list_records(rtype, name, None)
            identifier = resources[0]['id'] if resources else None

        LOGGER.debug('update_record: %s', identifier)

        if name:
            name = self._relative_name(name).lower()

        url = 'domains/{0}/records/{1}'.format(self.domain_id, identifier)
        self._put(url, data={
            'name': name.lower() if name else None,
            'type': rtype if rtype else None,
            'target': content if content else None
        })

        return True

    # Delete an existing record.
    # If record does not exist, do nothing.
    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        delete_resource_id = []
        if not identifier:
            resources = self._list_records(rtype, name, content)
            delete_resource_id = [resource['id'] for resource in resources]
        else:
            delete_resource_id.append(identifier)

        LOGGER.debug('delete_records: %s', delete_resource_id)

        for resource_id in delete_resource_id:
            self._delete('domains/{0}/records/{1}'.format(
                self.domain_id,
                resource_id
            ))

        return True

    # Helpers
    def _request(self, action='GET', url='', data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        default_headers = {
            'Accept': 'application/json',
            'Content-Type': 'application/json',
            'Authorization': 'Bearer {0}'.format(self._get_provider_option('auth_token'))
        }

        request_filter = query_params['filter'] if 'filter' in query_params else None
        if request_filter is not None:
            default_headers['X-Filter'] = json.dumps(request_filter)
            del query_params['filter']

        request_url = "{0}{1}".format(self.api_endpoint, url)

        response = requests.request(action, request_url, params=query_params,
                                    data=json.dumps(data),
                                    headers=default_headers)
        # if the request fails for any reason, throw an error.
        response.raise_for_status()
        if action == 'DELETE':
            return ''
        result = response.json()
        return result



"""Integration tests for DNSPod"""
from unittest import TestCase

import pytest
from lexicon.tests.providers.integration_tests import IntegrationTestsV1


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from define_tests.TheTests
# TODO: migrate to IntegrationTestsV2 and its extended test suite
class DnsPodProviderTests(TestCase, IntegrationTestsV1):
    """TestCase for DNSPod"""
    provider_name = 'dnspod'
    domain = 'capsulecd.com'

    def _filter_post_data_parameters(self):
        return ['login_token']

    # TODO: enable the skipped tests
    @pytest.mark.skip(reason="new test, missing recording")
    def test_provider_when_calling_update_record_should_modify_record_name_specified(self):
        return

    @pytest.mark.skip(reason="new test, missing recording")
    def test_provider_when_calling_list_records_after_setting_ttl(self):
        return



"""Module provider for auto"""
from __future__ import absolute_import
import argparse
import importlib
import logging
import pkgutil
import re
import subprocess

import six
import tldextract
from lexicon import providers
from lexicon.config import (
    ArgsConfigSource,
    ConfigResolver,
    legacy_config_resolver,
)


LOGGER = logging.getLogger(__name__)


def _get_available_providers():
    available_providers = {}
    for _, modname, _ in pkgutil.iter_modules(providers.__path__):
        if modname not in ('base', 'auto'):
            try:
                available_providers[modname] = importlib.import_module(
                    'lexicon.providers.' + modname)
            except ImportError:
                LOGGER.warning('Warning, the provider %s cannot be loaded due '
                               'to missing optional dependencies.', modname)

    return available_providers


AVAILABLE_PROVIDERS = _get_available_providers()


def _get_ns_records_domains_for_domain(domain):
    tlds = [tldextract.extract(ns_entry)
            for ns_entry in _get_ns_records_for_domain(domain)]

    return {'{0}.{1}'.format(tld.domain, tld.suffix) for tld in tlds}


def _get_ns_records_for_domain(domain):
    # Available both for Windows and Linux (if dnsutils is installed for the latter)
    try:
        output = subprocess.check_output(['nslookup', '-querytype=NS', domain + "."],
                                         stderr=subprocess.STDOUT, universal_newlines=True)
    except subprocess.CalledProcessError as error:
        if 'NXDOMAIN' in error.output:
            raise ValueError(
                'Error, domain {0} could not be resolved.'.format(domain))
        output = error.output

    pattern = re.compile(r'nameserver = (.*?)\.*\n')
    match = pattern.findall(output)

    if not match:
        raise ValueError('Error, could not find ns entries for domain {0}. '
                         'Does this domain is correctly configured ?'.format(domain))

    return match


def _relevant_provider_for_domain(domain):
    nameserver_domains = _get_ns_records_domains_for_domain(domain)
    relevant_providers = []

    for provider_name, provider_module in AVAILABLE_PROVIDERS.items():
        ns_domains = provider_module.NAMESERVER_DOMAINS

        # Test plain domain string comparison
        if {ns_domain for ns_domain in ns_domains
                if isinstance(ns_domain, six.string_types)} & nameserver_domains:
            relevant_providers.append((provider_name, provider_module))
            continue

        # Test domains regexp matching
        for ns_domain in ns_domains:
            if hasattr(ns_domain, 'match') \
                    and [nameserver_domain for nameserver_domain
                         in nameserver_domains if ns_domain.match(nameserver_domain)]:
                relevant_providers.append((provider_name, provider_module))
                continue

    if not relevant_providers:
        raise ValueError('Error, could not find the DNS provider for given domain {0}. '
                         'Found nameservers domains are {1}'.format(domain, nameserver_domains))

    if len(relevant_providers) > 1:
        LOGGER.warning('Warning, multiple DNS providers have been found for given domain %s, '
                       'first one will be used: %s\n'
                       'This may indicate a misconfiguration in one or more provider.',
                       domain, relevant_providers)

    return relevant_providers[0]


def provider_parser(subparser):
    """Configure provider parser for auto provider"""
    subparser.description = '''
        Provider 'auto' enables the Lexicon provider auto-discovery.
        Based on the nameservers declared for the given domain,
        Lexicon will try to find the DNS provider holding the DNS zone if it is supported.
        Actual DNS zone read/write operations will be delegated to the provider found:
        every environment variable or command line specific to this provider
        can be passed to Lexicon and will be processed accordingly.
        '''
    subparser.add_argument("--mapping-override", metavar="[DOMAIN]:[PROVIDER], ...",
                           help="comma separated list of elements in the form of "
                                "[DOMAIN]:[PROVIDER] to authoritatively map a "
                                "particular domain to a particular provider")

    # Explore and load the arguments available for every provider into the 'auto' provider.
    for provider_name, provider_module in AVAILABLE_PROVIDERS.items():
        parser = argparse.ArgumentParser(add_help=False)
        provider_module.provider_parser(parser)

        for action in parser._actions:  # pylint: disable=protected-access
            action.option_strings = [re.sub(
                r'^--(.*)$', r'--{0}-\1'.format(provider_name), option)
                                     for option in action.option_strings]
            action.dest = 'auto_{0}_{1}'.format(provider_name, action.dest)
            subparser._add_action(action)  # pylint: disable=protected-access

# Take care of the fact that this provider extends object, not BaseProvider !
# Indeed we want to delegate every parameter/method call to the delegate provider
# but __getattr__ is called only if the parameter/method cannot be found in the
# current Provider hierarchy. If it is object, it will be the case for every relevant
# call in the Lexicon library.
class Provider(object):  # pylint: disable=useless-object-inheritance
    """
    Implementation of the provider 'auto'.
    For the given domain, it will resolve the actual Provider class to use by inspecting the
    nameservers declared for the domain, using declared nameservers domain in each Lexicon DNS
    provider.
    Any call upon instantiation will be delegated to the resolved Provider.
    Any command line parameter or environment variable passed to Lexicon will be transfered to
    the resolved provider if it respect the naming convention: --[provider]-[parameter_name] for a
    command line parameter, or LEXICON_[PROVIDER]_PARAMETER_NAME for a environment variable.
    """
    def __init__(self, config):
        if not isinstance(config, ConfigResolver):
            # If config is a plain dict, we are in a legacy situation.
            # To protect the Provider API, the legacy dict is handled in a
            # correctly defined ConfigResolver.
            self.config = legacy_config_resolver(config)
        else:
            self.config = config

        self.domain = config.resolve('lexicon:domain')
        self.proxy_provider = None

    def authenticate(self):  # pylint: disable=too-many-locals
        """
        Launch the authentication process: for 'auto' provider, it means first to find the relevant
        provider, then call its authenticate() method. Almost every subsequent operation will then
        be delegated to that provider.
        """
        mapping_override = self.config.resolve('lexicon:auto:mapping_override')
        mapping_override_processed = {}
        if mapping_override:
            for one_mapping in mapping_override.split(','):
                one_mapping_processed = one_mapping.split(':')
                mapping_override_processed[one_mapping_processed[0]
                                           ] = one_mapping_processed[1]

        override_provider = mapping_override_processed.get(self.domain)
        if override_provider:
            provider = [
                element for element in AVAILABLE_PROVIDERS.items()
                if element[0] == override_provider][0]
            LOGGER.info('Provider authoritatively mapped for domain %s: %s.',
                        self.domain, provider.__name__)
            (provider_name, provider_module) = provider
        else:
            (provider_name, provider_module) = _relevant_provider_for_domain(self.domain)
            LOGGER.info('Provider discovered for domain %s: %s.',
                        self.domain, provider_name)

        new_config = ConfigResolver()
        new_config.with_dict({'provider_name': provider_name})

        target_prefix = 'auto_{0}_'.format(provider_name)
        for config_source in self.config._config_sources:  # pylint: disable=protected-access
            if not isinstance(config_source, ArgsConfigSource):
                new_config.with_config_source(config_source)
            else:
                # ArgsConfigSource needs to be reprocessed to rescope the provided
                # args to the delegate provider
                new_dict = {}
                for key, value in config_source._parameters.items():  # pylint: disable=protected-access
                    if key.startswith(target_prefix):
                        new_param_name = re.sub(
                            '^{0}'.format(target_prefix), '', key)
                        if provider_name not in new_dict:
                            new_dict[provider_name] = {}
                        new_dict[provider_name][new_param_name] = value
                    elif not key.startswith('auto_'):
                        new_dict[key] = value
                new_config.with_dict(new_dict)

        self.proxy_provider = provider_module.Provider(new_config)
        self.proxy_provider.authenticate()

    def __getattr__(self, attr_name):
        """
        Delegate any call to any parameter/method to the underlying provider.
        Method authenticate() must have been called before.
        """
        if not self.proxy_provider:
            raise ValueError('The \'auto\' provider requires its authenticate method '
                             'to be called before any subsequent parameter/method call.')

        return getattr(self.proxy_provider, attr_name)



"""Integration tests for CloudXNS"""
from unittest import TestCase

import pytest
from lexicon.tests.providers.integration_tests import IntegrationTestsV2


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from define_tests.TheTests


class CloudXNSProviderTests(TestCase, IntegrationTestsV2):
    """TestCase for CloudXNS"""
    provider_name = 'cloudxns'
    domain = 'capsulecd.com'

    def _filter_post_data_parameters(self):
        return ['login_token']

    # TODO: enable the skipped tests
    @pytest.mark.skip(reason="new test, missing recording")
    def test_provider_when_calling_update_record_should_modify_record_name_specified(self):
        return

    def _test_parameters_overrides(self):
        return {'api_endpoint': 'https://www.cloudxns.net/api2'}



"""Integration tests for Azure Cloud DNS"""
import re

from unittest import TestCase

from lexicon.tests.providers.integration_tests import IntegrationTestsV2


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from integration_tests.IntegrationTests


class AzureTests(TestCase, IntegrationTestsV2):
    """TestCase for Google Cloud DNS"""
    provider_name = 'azure'
    domain = 'full4ir.tk'

    def _test_parameters_overrides(self):
        return {'resource_group': 'dns-test'}

    def _filter_headers(self):
        return [('Authorization', 'Bearer TOKEN')]

    def _filter_post_data_parameters(self):
        return [('client_id', 'CLIENT_ID'), ('client_secret', 'CLIENT_SECRET')]

    def _filter_request(self, request):
        # Hide auth_tenant_id value in oauth token requests
        request.uri = re.sub(r'/[\w-]+/oauth2/token', '/TENANT_ID/oauth2/token',
                             request.uri)
        # Hide auth_subscription_id value in DNS requests
        request.uri = re.sub(r'/subscriptions/[\w-]+/', '/subscriptions/SUBSCRIPTION_ID/',
                             request.uri)
        return request

    def _filter_response(self, response):
        # Hide access_token value in oauth token responses
        response['body']['string'] = re.sub(
            br'"access_token":"[\w.-]+"', b'"access_token":"TOKEN"',
            response['body']['string'])
        response['body']['string'] = re.sub(
            br'\\/subscriptions\\/[\w-]+\\/', b'\\/subscriptions\\/SUBSCRIPTION_ID\\/',
            response['body']['string'])
        return response



"""test module for aliyun provider"""

# Test for one implementation of the interface
from unittest import TestCase
from lexicon.tests.providers.integration_tests import IntegrationTestsV2


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from integration_tests.IntegrationTests
class AliyunProviderTests(TestCase, IntegrationTestsV2):
    """Integration tests for Foo provider"""
    provider_name = 'aliyun'
    domain = 'mean.space'

    def _filter_query_parameters(self):
        """filter access key id && secret info"""
        return [('AccessKeyId', 'DUMMY_KEY_ID'),
                ('Signature', 'DUMMY_SIGNATURE'),
                ('SignatureNonce', 'DUMMY_SIGNATURE_NONCE'),
                ('Timestamp', 'DUMMY_TIMESTAMP')]



"""Integration tests for Cloudflare"""
from unittest import TestCase

from lexicon.tests.providers.integration_tests import IntegrationTestsV2


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from define_tests.TheTests
class CloudflareProviderTests(TestCase, IntegrationTestsV2):
    """TestCase for Cloudflare"""
    provider_name = 'cloudflare'
    domain = 'pacalis.net'

    def _filter_headers(self):
        return ['X-Auth-Email', 'X-Auth-Key', 'set-cookie']

    # We do not want to have "placeholder_auth_username" as default value for `--auth-username`
    # if Bearer tokens are used to execute the tests, because the non-emptiness of this flags
    # triggers interpretation of `--auth-key` as a Global API key.
    # Similarly for `--zone-id`, we want to control when its value is not empty, because
    # it will change the logic of the authentication process.
    def _test_fallback_fn(self):
        return lambda x: 'placeholder_' + x if x not in ('auth_username', 'zone_id') else ''



"""Integration tests for UltraDNS"""
from unittest import TestCase

from lexicon.tests.providers.integration_tests import IntegrationTestsV2


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from integration_tests.IntegrationTests
class UltradnsProviderTests(TestCase, IntegrationTestsV2):
    """TestCase for UltraDNS"""
    provider_name = 'ultradns'
    domain = 'example-abtest.com'

    def _filter_headers(self):
        return ['Authorization']

    def _filter_post_data_parameters(self):
        return ['username', 'password', 'accessToken']
